(*

                WinPrint v1.0       (c) 1992 Paul Bayliss
                Printer unit for Turbo Pascal for Windows

The major problem with TPW v1 is the total lack of information on how to
produce printed output. Since this is a rather important requirement of
many programs this lack is very annoying!!

I wrote WinPrint because of my own problems with printing from Windows.
The unit provides a number of routines to make printing easier. The
following gives a (very) brief outline of printing in Windows before
going on to explain how to use WinPrint.


PRINTING IN WINDOWS

Printing from Windows is not easy. Device independence means much more
work for the poor programmer - gone are the days of writeln(LST,'.....');

To print in Windows you send output to a device context, similar to
screen output. Getting the DC is rather complicated however since the
end user can change the printer at any time. This means that any
Windows Program must be able to identify the current printer at run time.

The current printer is listed in Win.INI undewr the [windows] section with
the key name of 'device' in the form:

             [windows]
             device=HP Laserjet III,HPPCL5A,LPT1:

The 3 sections, separated by a comma, refer to:

             1.  the printer name   ('HP Laserject III')
             2.  the printer driver ('HPPCLA5A')
             3.  the printer port   ('LPT1:')

All available printers are listed in the [devices] section of Win.ini:

              [devices]
              PCL / HP LaserJet=HPPCL,None
              HP LaserJet III=HPPCL5A,FILE:
              HP DeskJet Family=DESKJET,LPT1:
              Epson 24 pin=EPSON24,None

It is possible to change to any one of the available devices at any time.
Note that only one printer is connected to a port (eg LPT1:). The Windows
Control Panel refers to the connected printer as 'Active' and the device
with a port of 'None' as 'Inactive'. The ACTUAL printer port assignments
are listed in the [PrinterPorts] section of Win.ini

           [PrinterPorts]
           PCL / HP LaserJet=HPPCL,FILE:,15,45
           HP LaserJet III=HPPCL5A,FILE:,15,45
           HP DeskJet Family=DESKJET,LPT1:,15,900
           Epson 24 pin=EPSON24,LPT1:,15,900

To create the appropriate Device Context you must first get the current
(active) printer using GetProfileString. You then create a printer DC using
CreateDC:

    PrintDC:=CreateDC(Driver,DeviceName,Port,Nil);

This device context can then be used like a normal display DC (well almost)

When you print in Windows you are sending data to a virtual printer. The
GDI module create a metafile before sending it to the device driver
(eg HPPCL5A.DRV) which then translates the GDI code into code the printer
can understand. Device Drivers are DLL's and must be loaded using the
LoadLibrary() function.

Once the driver is loaded you use the 'Escape' function to control how the
printer works. The main commands are STARTDOC, NEWFRAME and ENDDOC.

Rather than produce an opus on printing in Windows I have tried to
annotate the code of WinPrint to explain what is happening as we go along.

If you want more information on printing in Windows try:

   1. Programming Windows v3   Charles Petzold
   2. Microsoft Windows Programmer's Reference
   3. Microsoft Windows Guide to Programming

All these books use the 'C' language but most the information is
understandable.


USING WINPRINT

Winprint provides several routines which should help make printing
easier:

  * Procedure SetupPrinter(Window : Hwnd);

    Allows you to display the 'Setup' dialog of the current printer
    driver so that the settings can be altered. Pass the current
    window handle (Hwindow) to the routine.

  * Procedure SelectPrinter(Window : Hwnd);

    Displays a list of all available printers, allowing you to change
    the 'active' printer. Pass the current window handle (Hwindow)
    to the routine.

  * Procedure StartPrinter(Qname : Pchar);

    This routine loads the active printer driver, loads Print Manager,
    gets the PrinterDC (PRINTDC) and displays the printer abort dialog
    box. 'QName' is the text to display in the Print Manager window.
    You should always supply a Qname to ensure that multi-page documents
    print correctly. Once you have called StartPrinter you can then
    use PRINTDC as you would any other device context,

    eg Rectangle(PRINTDC,10,10,100,150)
       Textout(PRINTDC,10,10,'This is a line of text',22);

    You can use GetDeviceCaps to find the size of a page.

  * Procedure NewPage;

    Sends a NEWFRAME escape to the driver, starting a newpage.
    Note: You MUST issue at least one NewPage command for Windows
    to print anything.

  * Procedure StopPrinter;

    Ends printing, removes the driver from memory and destoys the
    printer Device Context. Removes to print dialog box.

  * Procedure DisplayPrintCompleted(percent : byte);

    Allows you to display the percentage of the printing that
    as been completed in the printer dialog box.

  * Procedure DisplayPrintInfo(Info : Pchar);

    Allows you to display information about the current print job in
    the printer dialog window


To print anything using WinPrint the basic procedure is as follows

   StartPrinter(Hwindow);
   .
   .
   Information to Print
   .
   .
   NewPage;
   StopPrinter;

If you wish to print more than one page simple repeat the Newpage,
information, NewPage loop until you have finished. You MUST have at least
one NewPage.


WinPrint also defines a new object, TLinePrinter, which simplfies the
job of printing simple text. TLinePrinter maintains a record of the current
position on the paper, moving to the next line once the edge of the paper
as been reached, moving down the paper by the height of the current font.
Once the bottom of the paer as been reached a formfeed is issued and
printing continues on the next page.

To use the new object simply initalise a new object and then call the
appropriate method:

      * Init(Qname : Pchar);

        Initialises the object. Qname is the same as for Startprinter.

      * Write(data : Pchar)

        Works like the normal TP Write() procedure. Printing starts at
        the current cursor position. Does not perform a line feed.

      * Writeln(data : Pchar)

        Works like the normal Writeln() procedure. Printing starts at
        the current cursor position. Performs a line feed once the data
        as been printed.

      * NewPage

        Allows you to manually issue a formfeed. Moves the current cursor
        position to the top left of the page.

      * NewLine

        Allows you to issue a linefeed. Moves the current cursor position
        to the left of the page, on the next line down.

      * Done

        Destroys the object after removing the printer driver.
        Same as StopPrinter.


EXAMPLE

Uses Wobjects,WinPrint;

Var
 LinePrinter : PLinePrinter;

 begin
   LinePrinter:=new(PLinePrinter,init('Line Printer Test'));
   { Initalise the object }
   for x:=1 to 20 do
   begin
     LinePrinter^.writeln('Test Information');
   end;
   { Print 20 lines of 'Test Information' }
   LinePrinter^.newpage;
   { Issue formfeed - MUST }
   LinePrinter^.done;
   { Stop printing }
 end;


GENERAL INFO

This is my first stab at printing with TPW. AS such I cannot promise that
it will work in every situation and I will not accept any liabilty etc
for ANY damage caused by using these routines. However - it works for me!

Although I wish to keep copyright to WinPrint I offer it free to anyone
who can use it. I would like to be notified of any bugs or enhancements
that you may make : Compuserve [70374,2414]
=========================================================================
*)

Unit WinPrint;

INTERFACE

uses WObjects,WinTypes,WinProcs,Strings;

Var
   PrintDC : hDC;
   PrintingAborted : Boolean;

Type
   PLinePrinter = ^TLinePrinter;
   TLinePrinter = object(tobject)
    LineHeight,MaxX,MaxY,CurX,CurY : word;
    Metrics : tTextmetric;
    Constructor Init(Qname : Pchar);
    Destructor Done; virtual;
    Procedure NewLine; virtual;
    Procedure NewPage; virtual;
    Function LineWidth(data : pchar) : integer; virtual;
    Procedure Write(Data : Pchar); virtual;
    Procedure WriteLn(data : Pchar); virtual;
   end;


Procedure SetupPrinter(Window : Hwnd);
Procedure SelectPrinter(Window : Hwnd);
Procedure StartPrinter(Qname : Pchar);
Procedure NewPage;
Procedure StopPrinter;
Procedure DisplayPrintCompleted(percent : byte);

IMPLEMENTATION
{$R WinPrint.RES}


Type
    (* Windows 3 Configuration Function *)
    TPGetExtDevMode = Function(Window  : Hwnd;
                               Handle  : tHandle;
                               OutMode : tDevMode;
                               DevName : Pchar;
                               Output  : Pchar;
                               InMode  : TDevMode;
                               Profile : Pchar;
                               PMode   : word) : Boolean;

    (* Windows 2 Configuration Function *)
    TPGetDevMode    = Function(Window  : Hwnd;
                               Handle  : tHandle;
                               DevName : Pchar;
                               OutPut  : Pchar) : Boolean;

  ErrorCode = (NoPrinterInfo,NoDC,NoMemory);

  PPrintAbortDlg = ^TPrintAbortDlg;
  TPrintAbortDlg = object(TDialog)
   Procedure Cancel(var MSG : TMessage); virtual id_first + id_cancel;
  end;

PProfileArray = ^TProfileArray;
TProfileArray = object(TSortedCollection)
 Function Compare(Key1,Key2 : Pointer) : Integer; virtual;
end;

Function TProfileArray.Compare(Key1,Key2 : Pointer): integer;
begin
  Compare:=-1;
end;

Var
   DevMode : tDevMode;
   pAddr   : tFarProc;
   DevHandle : tHandle;
   ProfileString,Message : array[0..255] of char;
   Driver,DriverName,DeviceName,Port : array[0..79] of char;
   DevName   : Pchar;
   ChPtr,ChPos : Pchar;
   Count : Longint;
   ProfileArray : PSortedCollection;
   PrintDialog : PDialog;
   PrintAbortDlg : PPrintAbortDlg;
   lpPrintAbortProc : tFarProc;

   Tempstr : array[0..20] of char;

Procedure ParseStr(Pstr : pchar; Count : longint; ch : char;
                   var ReturnedArray : PsortedCollection);
var
   Ploop,Pindex : longint;
   TempWord : array[0..80] of char;
begin
     TempWord[0]:=#0;
     Pindex:=0;
     for PLoop:=0 to Count-1 do
     begin
          if Pstr[Ploop]<>ch then
          begin
            TempWord[Pindex]:=Pstr[Ploop];
            inc(Pindex);
            TempWord[Pindex]:=#0;
          end
          else
          begin
            ReturnedArray^.insert(strnew(TempWord));
            TempWord[0]:=#0;
            Pindex:=0;
          end;
     end;
     if (ch<>#0) then ReturnedArray^.insert(strnew(TempWord));
end;

Function ParseProfileString(AppName,KeyName,Default : pchar;
                            var ReturnedArray : PSortedCollection) : integer;
begin
  Count:=GetProfileString(AppName,KeyName,Default,ProfileString,sizeof(ProfileString));
  if KeyName=NIL then
    ParseStr(ProfileString,Count,#0,ReturnedArray)
  else
    ParseStr(ProfileString,Count,',',ReturnedArray);
  ParseProfileString:=ReturnedArray^.Count;
end;


Function GetPrinterInfo : Boolean;
begin
  GetPrinterInfo:=FALSE;
  ProfileArray:=new(PProfileArray,init(1,1));
  if ParseProfileString('Windows','Device','NONE',ProfileArray)<3 then exit;
  StrCopy(DeviceName,ProfileArray^.at(0));
  strcopy(DriverName,ProfileArray^.at(1));
  StrCopy(Port,ProfileArray^.at(2));
  dispose(ProfileArray);
  strcopy(Driver,DriverName);
  strcat(DriverName,'.DRV');
  GetPrinterInfo:=TRUE;
end;


Procedure SetupPrinter(Window : Hwnd);
begin
  if Not GetPrinterInfo then
  begin
    Messagebox(Window,'Unable to Load Printer Configuration Routine',
               'ERROR',mb_iconexclamation or mb_ok);
    exit;
  end;
  DevHandle:=LoadLibrary(Drivername);
  Paddr:=getprocaddress(DevHandle,'ExtDeviceMode');
  if (Paddr<>NIL) then
  begin
    TPGetExtDevMode(Paddr) (window,DevHandle,DevMode,DriverName,Port,DevMode,NIL, dm_update or dm_prompt);
  end
  else
  begin
    Paddr:=getprocaddress(DevHandle,'DeviceMode');
    if (Paddr<>NIL) then
    begin
      TPGetDevMode(Paddr) (window,DevHandle,DriverName,Port);
    end;
  end;
  FreeLibrary(DevHandle);
end;


Type
    PPrintDialog = ^TPrintDialog;
    TPrintDialog = object(TDialog)
     DefaultPrinter : array[0..80] of char;
     DefaultDriver : array[0..12] of char;
     DefaultPort  : array[0..5] of char;
     PrintArray : PSortedCollection;
     Procedure Setupwindow; virtual;
     Procedure AvailableList(var MSG : TMessage); virtual id_first + 103;
     Procedure OK(var MSG : TMessage); virtual id_first + id_ok;
     Procedure cancel(var MSG : TMessage); virtual id_first + id_cancel;
     Procedure Setup(var MSG : TMessage); virtual id_first + 105;
    end;

    Procedure TPrintDialog.setup(var MSG : TMessage);
    begin
      setupprinter(hwindow);
    end;

    Procedure TPrintDialog.setupwindow;
    const TabArray : array[1..2] of integer = (30,150);
    var PrnCount,PrnIndex : integer;
    begin
      TDialog.Setupwindow;
      senddlgitemmessage(hwindow,103,lb_settabstops,2,longint(@TabArray));
      ProfileArray:=new(PProfileArray,init(1,1));
      if not ParseProfileString('Windows','Device','NONE',ProfileArray)=3
      then enddlg(2);
      strcopy(DefaultPrinter,ProfileArray^.at(0));
      StrCopy(DefaultDriver,ProfileArray^.at(1));
      StrCopy(DefaultPort,ProfileArray^.at(2));
      strEcopy(StrEcopy(StrEcopy(Message,ProfileArray^.at(0)),
                                         ',  Port='),
                                         ProfileArray^.at(2));
      setdlgitemtext(hwindow,102,Message);
      Dispose(ProfileArray);
      ProfileArray:=new(PProfileArray,init(1,1));
      for PrnCount:=0 to ParseProfileString('Devices',nil,'NONE',ProfileArray)-1 do
      begin
        PrintArray:=new(PProfileArray,init(1,1));
        ParseProfileString('PrinterPorts',ProfileArray^.at(Prncount),'NONE Found',PrintArray);
        StrEcopy(StrECopy(StrEcopy(Message,ProfileArray^.at(PrnCount)),
                                           ',  Port='),
                                           PrintArray^.at(1));
        senddlgitemmessage(hwindow,103,lb_addstring,0,longint(@message));
        Dispose(PrintArray);
      end;
      Dispose(ProfileArray);
    end;

    Procedure TPrintDialog.AvailableList(var MSG : TMessage);
    var PrnIndex : integer;
    begin
      Case MSG.LParamHI of
        lbn_SelChange : Begin
                        PrnIndex:=senddlgitemmessage(Hwindow,103,lb_getcursel,0,0);
                        senddlgitemmessage(hwindow,103,lb_gettext,prnindex,longint(@message));
                        setdlgitemtext(hwindow,102,message);
                        end;
        lbn_DblClk    : begin
                        PrnIndex:=senddlgitemmessage(Hwindow,103,lb_getcursel,0,0);
                        senddlgitemmessage(hwindow,103,lb_gettext,prnindex,longint(@message));
                        setdlgitemtext(hwindow,102,message);
                        OK(MSG);
                        end;
      end;
    end;

    Procedure TPrintDialog.OK(var MSG : TMessage);
    var Mindex,MPos : Pchar;
        Printer : array[0..80] of char;
        Driver,port : array[0..12] of char;
    begin
      getdlgitemtext(hwindow,102,message,80);
      MIndex:=message;
      MPos:=StrScan(Mindex,',');
      strLcopy(Printer,Mindex,(MPos-Mindex));
      ProfileArray:=new(PProfileArray,init(1,1));
      ParseProfileString('PrinterPorts',Printer,nil,ProfileArray);
      strcopy(Driver,profilearray^.at(0));
      strcopy(Port,ProfileArray^.at(1));
      dispose(ProfileArray);

      StrEcopy(StrEcopy(StrEcopy(strEcopy(strEcopy(
          message,Printer),','),Driver),','),Port);
      Writeprofilestring('Windows','Device',message);

      strecopy(strecopy(strecopy(Message,Driver),','),Port);
      Writeprofilestring('Devices',Printer,message);

      strecopy(strecopy(strecopy(Message,DefaultDriver),','),'None');
      Writeprofilestring('Devices',DefaultPrinter,message);

      strcopy(message,'windows');
      sendmessage($FFFF,wm_wininichange,0,longint(@message));
      strcopy(message,'Devices');
      sendmessage($FFFF,wm_wininichange,0,longint(@message));
      TDialog.OK(MSG);
    end;

    Procedure TPrintDialog.Cancel(var MSG : TMessage);
    begin
      TDialog.Cancel(MSG);
    end;


Procedure SelectPrinter(Window : Hwnd);
begin
  PrintDialog:=new(PPrintDialog,init(Application^.Mainwindow,'SelectPrinter'));
  Application^.execdialog(PrintDialog);
  setfocus(window);
end;

(*=====================================================================*)




  Procedure TPrintAbortDlg.Cancel(var MSG : TMessage);
  begin
     PrintingAborted:=TRUE;
     setwindowtext(hwindow,'Printing Cancelled...');
  end;


Procedure PrinterError(Error : integer);
var Message : array[0..80] of char;
begin
  case error of
    sp_Error : strcopy(message,'Unknown (general) error');
    sp_OutofDisk : strcopy(message,'Insufficient disk space for printing');
    sp_outofmemory : strcopy(message,'insufficient memory for printing');
  end;
  Messagebox(getfocus,message,'Printer Error',mb_iconstop or mb_ok);
end;

Function PrintAbortProc(PrnDC : hdc; code : word) : Boolean; EXPORT;
var WinMSG : tMSG;
begin
  if (code<>0) and (Code<>sp_Userabort) then PrinterError(Code);
  While not PrintingAborted and Peekmessage(WinMSG,0,0,0,pm_remove) do
  begin
    if not Isdialogmessage(PrintAbortDlg^.Hwindow,WinMSG) then
    begin
      Translatemessage(WinMSG);
      Dispatchmessage(WinMSG);
    end;
  end;
  PrintAbortProc:=not PrintingAborted;
end;

Procedure StartPrinter(QName : Pchar);
begin
  if not GetPrinterInfo then exit;
  PrintAbortDlg:=new(PPrintAbortDlg,init(Application^.mainwindow,'AbortPrint'));
  PrintAbortDlg^.create;
  PrintDC:=CreateDC(Driver,DeviceName,Port,Nil);
  if printdc=0 then messagebox(getfocus,'','No DC',0);
  lpPrintAbortProc:=makeprocinstance(@PrintAbortProc,hinstance);
  if Escape(PrintDC,SetAbortProc,0,lpPrintAbortProc,nil)<1 then
  begin
     messagebox(getfocus,'Unable to install Printer Procedure',
                'Printer Error',0);
     PrintAbortDlg^.destroy;
     freeprocinstance(lpPrintabortproc);
     deletedc(PrintDC);
     exit;
  end;
  Escape(PrintDC,STARTDOC,sizeof(Qname),Qname,NIL);
  PrintAbortDlg^.show(sw_normal);
end;

Procedure NewPage;
begin
   setwindowtext(PrintAbortDlg^.Hwindow,'Spooling, Please wait');
   Escape(PrintDC,NEWFRAME,0,nil,nil);
   setwindowtext(PrintAbortDlg^.Hwindow,'Printing');
end;

Procedure StopPrinter;
begin
   Escape(PrintDC,ENDDOC,0,nil,nil);
   PrintAbortDlg^.destroy;
   freeprocinstance(lpPrintabortproc);
   deletedc(PrintDC);
end;

Procedure DisplayPrintCompleted(percent : byte);
var tempstr : array[0..20] of char;
begin
   str(Percent,tempstr);
   strcat(tempstr,'% Complete');
   setdlgitemtext(PrintAbortDlg^.Hwindow,4,tempstr);
end;

Procedure DisplayPrintInfo(Info : Pchar);
begin
   setdlgitemtext(PrintAbortDlg^.hwindow,3,info);
end;

(*====== Line Printer Methods ========================================*)

Constructor TLinePrinter.Init(Qname : Pchar);
begin
  TObject.Init;
  StartPrinter(Qname);
  gettextmetrics(PrintDC,metrics);
  LineHeight:=Metrics.tmheight+metrics.tmExternalLeading;
  MaxX:=getdevicecaps(PrintDC,Horzres);
  MaxY:=Getdevicecaps(PrintDC,VertRes);
  CurX:=0; CurY:=0;
end;

Destructor TLinePrinter.Done;
begin
  TObject.done;
  stopprinter;
end;

Procedure TLinePrinter.NewLine;
begin
  CurX:=0;
  inc(CurY,LineHeight);
  if CurY>MaxY then newpage;
end;

Function TLinePrinter.LineWidth(data : Pchar) : integer;
begin
   if data<>NIL then
     LineWidth:=(lo(getTextExtent(PrintDC,data,strlen(data))))
   else
     LineWidth:=0;
end;

Procedure TLinePrinter.Write(data : Pchar);
begin
  if (CurX + LineWidth(data))> MaxX then newline;
  Textout(PrintDC,CurX,CurY,data,strlen(data));
  inc(CurX,linewidth(data));
end;

Procedure TLinePrinter.WriteLn(data : Pchar);
begin
  Write(Data);
  newline;
end;

Procedure TLinePrinter.NewPage;
begin
 setwindowtext(PrintAbortDlg^.Hwindow,'Spooling, Please wait');
 Escape(PrintDC,NEWFRAME,0,nil,nil);
 setwindowtext(PrintAbortDlg^.Hwindow,'Printing');
 CurX:=0; CurY:=0;
end;


end.
